﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace _485HIDApp
{
    public partial class Form1 : Form
    {
        bool keepReading = true;

        string vid = "vid_277d";
        string pid = "pid_7002";

        [DllImport("hid.dll", EntryPoint = "HidD_GetHidGuid", SetLastError = true)]
        static extern void HidD_GetHidGuid(ref Guid hidGuid);

        [DllImport("setupapi.dll", CharSet = CharSet.Auto, SetLastError = true)]
        static extern IntPtr SetupDiGetClassDevs(ref Guid ClassGuid, IntPtr Enumerator, IntPtr hwndParent, int Flags);

        [StructLayout(LayoutKind.Sequential)]
        public struct SECURITY_ATTRIBUTES
        {
            public int length;
            public int securityDescriptor;
            public int inheritHandle;
        }

        [DllImport("setupapi.dll", SetLastError = true)]
        protected static extern bool SetupDiEnumDeviceInterfaces(IntPtr hDevInfo, IntPtr devInfo, ref Guid interfaceClassGuid, UInt32 memberIndex, ref SP_DEVICE_INTERFACE_DATA deviceInterfaceData);

        [StructLayout(LayoutKind.Sequential)]
        public struct SP_DEVICE_INTERFACE_DATA
        {
            public Int32 size;
            public Guid interfaceClassGuid;
            public Int32 flags;
            private UIntPtr reserved;
        }

        [DllImport(@"setupapi.dll", CharSet = CharSet.Auto, SetLastError = true)]
        static extern bool SetupDiGetDeviceInterfaceDetail(IntPtr DeviceInfoSet, ref SP_DEVICE_INTERFACE_DATA deviceInterfaceData, IntPtr aPtr, UInt32 deviceInterfaceDetailDataSize, ref UInt32 requiredSize, IntPtr bPtr);

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
        public struct SP_DEVICE_INTERFACE_DETAIL_DATA
        {
            public Int32 cbSize;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 256)]
            public string DevicePath;
        }

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern int CreateFile(string lpFileName, uint dwDesiredAccess, uint dwShareMode, uint lpSecurityAttributes, uint dwCreationDisposition, uint dwFlagsAndAttributes, uint hTemplateFile);

        /*.[StructLayout(LayoutKind.Sequential)]
        public struct HIDD_ATTRIBUTES
        {
            public int size;
            public System.UInt16 vid;
            public System.UInt16 pid;
            public System.UInt16 version;
        }

        [DllImport("hid.dll", SetLastError = true)]
        private static extern int HidD_GetAttributes(int handle, ref HIDD_ATTRIBUTES attributes);*/

        [DllImport("hid.dll")]
        public static extern bool HidD_GetSerialNumberString(int handle, IntPtr buffer, int bufferLength);

        [DllImport("kernel32.dll")]
        static public extern int WriteFile(int handle, ref byte buffer, int numberOfBytesToWrite, ref int numberOfBytesWritten, int overlapped);

        [DllImport("kernel32.dll")]
        static public extern int ReadFile(int handle, byte[] buffer, int numberOfBytesToRead, ref int numberOfBytesRead, ref NativeOverlapped overlapped);

        [DllImport("hid.dll", SetLastError = true)]
        private static extern bool HidD_GetProductString(int handle, IntPtr buffer, uint bufferLength);

        [DllImport("hid.dll", CharSet = CharSet.Auto, SetLastError = true)]
        static extern Boolean HidD_GetManufacturerString(int handle, IntPtr buffer, uint bufferLength);

        [DllImport("kernel32.dll")]
        static public extern int WaitForSingleObject(int handle, int milliseconds);

        [DllImport("kernel32.dll", SetLastError = true)]
        static extern bool GetOverlappedResult(int handle, [In] ref NativeOverlapped overlapped, out uint numberOfBytesTransferred, bool wait);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto)]
        static public extern int CreateEvent(ref SECURITY_ATTRIBUTES securityAttributes, int manualReset, int initialState, string name);

        [DllImport("kernel32.dll")]
        static public extern int CancelIo(int handle);

        [DllImport("kernel32.dll")]
        static public extern int CancelIoEx(int handle, /*NativeOverlapped*/IntPtr overlapped);

        List<Device> devices = new List<Device>();
        Device currentDevice = null;
        public static NativeOverlapped HIDOverlapped;
        Thread readThread;
        int eventNum = 0;
        SECURITY_ATTRIBUTES securityAttributes = new SECURITY_ATTRIBUTES();
        delegate void ThreadSafeAddConsoleTextCallBack(string text);

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            refreshPrinters();
        }

        public void startReading()
        {
            readThread = new Thread(readThreadEntry);
            readThread.Start();
        }

        public void refreshPrinters()
        {
            hidDevicesDropDown.Items.Clear();
            Guid guid = new Guid();
            HidD_GetHidGuid(ref guid);
            IntPtr devInfoPtr = SetupDiGetClassDevs(ref guid, IntPtr.Zero, IntPtr.Zero, 0x12);
            int devInfoInt = (int)Marshal.PtrToStructure(devInfoPtr, typeof(int));
            securityAttributes.securityDescriptor = 0;
            securityAttributes.inheritHandle = System.Convert.ToInt32(true);
            securityAttributes.length = Marshal.SizeOf(securityAttributes);
            SP_DEVICE_INTERFACE_DATA deviceInterfaceData = new SP_DEVICE_INTERFACE_DATA();
            deviceInterfaceData.size = Marshal.SizeOf(deviceInterfaceData);
            for (uint i = 0; SetupDiEnumDeviceInterfaces(devInfoPtr, IntPtr.Zero, ref guid, i, ref deviceInterfaceData); i++)
            {
                SP_DEVICE_INTERFACE_DETAIL_DATA spDeviceInterfaceDetailData = new SP_DEVICE_INTERFACE_DETAIL_DATA();
                uint requiredSize = 0;
                SetupDiGetDeviceInterfaceDetail(devInfoPtr, ref deviceInterfaceData, IntPtr.Zero, 0, ref requiredSize, IntPtr.Zero);
                uint size = requiredSize;
                if (IntPtr.Size == 8) // for 64 bit operating systems
                    spDeviceInterfaceDetailData.cbSize = 8;
                else
                    spDeviceInterfaceDetailData.cbSize = 4 + Marshal.SystemDefaultCharSize; // for 32 bit systems
                IntPtr detailDataPtr = Marshal.AllocHGlobal(Marshal.SizeOf(spDeviceInterfaceDetailData));
                Marshal.StructureToPtr(spDeviceInterfaceDetailData, detailDataPtr, false);
                SetupDiGetDeviceInterfaceDetail(devInfoPtr, ref deviceInterfaceData, detailDataPtr, size, ref requiredSize, IntPtr.Zero);
                spDeviceInterfaceDetailData = (SP_DEVICE_INTERFACE_DETAIL_DATA)Marshal.PtrToStructure(detailDataPtr, typeof(SP_DEVICE_INTERFACE_DETAIL_DATA));
                string devicePath = spDeviceInterfaceDetailData.DevicePath;
                int handle = CreateFile(devicePath, 0xC0000000/*read and write*/, 0x3/*file share read and write*/, 0, 3/*open existing*/, 0, 0);//although this is specifying read and write this creates the first handle which will be used for writing
                IntPtr buffer = Marshal.AllocHGlobal(100);
                HidD_GetSerialNumberString(handle, buffer, 100);
                string serialNumberString = Marshal.PtrToStringAuto(buffer);
                buffer = Marshal.AllocHGlobal(100);
                HidD_GetProductString(handle, buffer, 100);
                string productString = Marshal.PtrToStringAuto(buffer);
                buffer = Marshal.AllocHGlobal(100);
                HidD_GetManufacturerString(handle, buffer, 100);
                string manufacturerString = Marshal.PtrToStringAuto(buffer);
                int readHandle = CreateFile(devicePath, 0xC0000000/*read and write*/, 0x3/*file share read and write*/, 0, 3/*open existing*/, 0, 0);//although this is specifying read and write this creates the first handle which will be used for reading
                Device d = new Device(devicePath, handle, readHandle, manufacturerString, productString, serialNumberString);
                devices.Add(d);
                hidDevicesDropDown.Items.Add(d.manufacturer + " - " + d.product + " - " + d.serialNumber);
                if (devicePath.Contains(vid) && devicePath.Contains(pid))
                    hidDevicesDropDown.SelectedIndex = (int)i;
            }
        }

        public void sendText()
        {
            if (sendTextBox.Text.Length > 0)
            {
                string asciiText = sendTextBox.Text + Environment.NewLine;
                if (hexRadioButton.Checked)
                {
                    string[] newLines = new string[sendTextBox.Lines.Length];
                    for (int i = 0; i < sendTextBox.Lines.Length; i++)
                        newLines[i] = toHex(sendTextBox.Lines[i]);
                    sendTextBox.Lines = newLines;
                }
                consoleTextBox.Text += sendTextBox.Text + Environment.NewLine;
                consoleTextBox.SelectionStart = consoleTextBox.Text.Length;
                consoleTextBox.ScrollToCaret();
                sendTextBox.Text = "";
                char[] asciiChars = asciiText.ToCharArray();
                byte[] outputBuffer = new byte[asciiChars.Length];
                for (int i = 0; i < outputBuffer.Length; i++)
                    outputBuffer[i] = (byte)asciiChars[i]; //copy the ascii character array into the output buffer
                sendBytes(outputBuffer);
            }
        }

        public void sendBytes(byte[] outputBuffer)
        {
            int written = 0;
            byte[] dataBuffer = new byte[33];
            int repCount = (outputBuffer.Length / (dataBuffer.Length - 1)); // number of reports to send is total data size/report size
            if ((outputBuffer.Length % (dataBuffer.Length - 1) != 0)) // there is some "extra" data that doesn't fill a whole report
                repCount++; //add another report and pad it with zeros
            for (int j = 0; j < repCount; j++) //loop through the reports to send
            {
                dataBuffer = new byte[33];
                int outputOffset = j * 32;
                for (int i = 1; i < dataBuffer.Length && outputOffset + i - 1 < outputBuffer.Length; i++) //fill dataBuf with 32 values from outputBuffer then write
                    dataBuffer[i] = outputBuffer[outputOffset + i - 1];
                WriteFile(currentDevice.handle, ref dataBuffer[0], dataBuffer.Length, ref written, 0);
            }
        }

        public void threadSafeAddConsoleText(string text)
        {
            if (this.consoleTextBox.InvokeRequired)
            {
                ThreadSafeAddConsoleTextCallBack d = new ThreadSafeAddConsoleTextCallBack(threadSafeAddConsoleText);
                this.Invoke(d, new object[] { text });
            }
            else
            {
                if (hexRadioButton.Checked)
                    consoleTextBox.Text += toHex(text);
                else
                    consoleTextBox.Text += text;
            }
        }

        public void readThreadEntry()
        {
            if (eventNum == 0)
                eventNum = CreateEvent(ref securityAttributes, 1, 1, "");
            HIDOverlapped.OffsetLow = 0;
            HIDOverlapped.OffsetHigh = 0;
            HIDOverlapped.EventHandle = (IntPtr)eventNum;
            int lastReadResult = 0;
            int readResult = 0;
            while (keepReading)
            {
                int read = 0;
                byte[] buffer = new byte[33];
                try
                {
                    lastReadResult = readResult;
                    ReadFile(currentDevice.readHandle, buffer, 33, ref read, ref HIDOverlapped);
                    readResult = WaitForSingleObject(currentDevice.readHandle, 1000);
                    if (readResult == 0)
                    {
                        string response = "";
                        for (int i = 0; i < buffer.Length && i < read; i++)
                            response += (char)buffer[i];
                        if (response.Length > 0)
                        {
                            response = response.Replace(((char)0).ToString(), "");
                            threadSafeAddConsoleText(response);
                            if (response[response.Length - 1] == (char)0)
                                threadSafeAddConsoleText(Environment.NewLine);
                            uint count = (uint)HIDOverlapped.InternalHigh;
                            bool success = GetOverlappedResult(currentDevice.readHandle, ref HIDOverlapped, out count, true);
                        }
                    }
                    else if (readResult == 0x102)
                        CancelIo(currentDevice.readHandle);
                }
                catch (ThreadAbortException tae) { }
                catch (Exception ex) { }
            }
            keepReading = true;//for the next read thread
        }

        private void refreshButton_Click(object sender, EventArgs e)
        {
            refreshPrinters();
        }

        private void sendTextBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)13)
                sendText();
        }

        private void hidDevicesDropDown_SelectedIndexChanged(object sender, EventArgs e)
        {
            try
            {
                if (readThread != null)
                {
                    keepReading = false;
                    int cancelResult = CancelIoEx(currentDevice.readHandle, IntPtr.Zero);//cancel current read
                    cancelResult = CancelIoEx(currentDevice.handle, IntPtr.Zero);//cancel current write
                    readThread.Join();
                }
            }
            catch (Exception ex) { }
            currentDevice = devices[hidDevicesDropDown.SelectedIndex];
            label4.Text = hidDevicesDropDown.Text;
            startReading();
        }

        private void sendFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog fd = new OpenFileDialog();
            if (fd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                byte[] fileBytes = File.ReadAllBytes(fd.FileName);
                sendBytes(fileBytes);
            }
        }

        private void asciiRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            string[] newLines;
            if (asciiRadioButton.Checked)
            {
                newLines = new string[consoleTextBox.Lines.Length];
                for (int i = 0; i < consoleTextBox.Lines.Length; i++)
                    newLines[i] = toAscii(consoleTextBox.Lines[i]);
                consoleTextBox.Lines = newLines;
            }
            else
            {
                newLines = new string[consoleTextBox.Lines.Length];
                for (int i = 0; i < consoleTextBox.Lines.Length; i++)
                    newLines[i] = toHex(consoleTextBox.Lines[i]);
                consoleTextBox.Lines = newLines;
            }
        }

        public string toHex(string ascii)
        {
            char[] chars = ascii.ToCharArray();
            string newStr = "";
            for (int i = 0; i < chars.Length; i++)
            {
                if ((int)chars[i] < 16)
                    newStr += "0";
                newStr += String.Format("{0:X}", (int)chars[i]) + " ";
            }
            return newStr;
        }

        public string toAscii(string hex/*space separated*/)
        {
            string newStr = "";
            for (int i = 0; i < hex.Length; i += 3)
                newStr += Convert.ToChar(Convert.ToUInt32(hex.Substring(i, 2), 16));
            return newStr;
        }

        private void sendTextButton_Click(object sender, EventArgs e)
        {
            sendText();
        }

        private void sendTextBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.V && e.Modifiers == Keys.Control)
                sendTextBox.Text = Clipboard.GetText();
        }


        const int MAX_TTF_SIZE = 524288;

        private void sendTTFFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog fd = new OpenFileDialog();
            if (fd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                byte[] fileBytes = File.ReadAllBytes(fd.FileName);

                // needed for FGL ttf download
                int ttfFileSize = fileBytes.Length;

                // verify that the file size does not exceed the max ttf size for the printer
                if (ttfFileSize <= MAX_TTF_SIZE)
                {
                    // build the fgl command to send before the ttf data is sent
                    string fglTTFCommandString = "<ttf" + ttfFileSize.ToString() + ">";

                    char[] fglTTFCommandAsCharArray = fglTTFCommandString.ToCharArray();
                    byte[] outputBuffer = new byte[fglTTFCommandAsCharArray.Length];
                    for (int i = 0; i < outputBuffer.Length; i++)
                        outputBuffer[i] = (byte)fglTTFCommandAsCharArray[i]; //copy the ascii character array into the output buffer
                    sendBytes(outputBuffer); // send the fgl command

                    sendBytes(fileBytes); // ... then send the ttf file
                }
                else // the ttf file is too large for the printer to handle
                {
                    MessageBox.Show("The TrueType Font file exceeds the maximum of 512 KB.\n", "Error downloading file...", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }

                
            }
        }
    }
}
